/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASSEMBLYJOBSUBSET_CPP
#define ASSEMBLYJOBSUBSET_CPP


#include "AssemblyJobSubset.h"
#include "ScoredSeqFlip.h"
#include "AssemblyException.h"
#include "Alignment.h"
#include <queue>
#include <string>
#include <iostream>
#include <typeinfo>
#include <algorithm>
#include <omp.h>
#include "MatchSeqTest.h"
using namespace::std;

AssemblyJobSubset::AssemblyJobSubset(){}
AssemblyJobSubset::AssemblyJobSubset(set<ScoredSeq*>* seqs,
				     float minFractId,
				     AssemblyJob::AssemblyStrandedness strandedness) :
  _wasRun(false),
  _inputRemoved(false),
  _novelRemoved(false),
  _strandedness(strandedness),
  _minFractId(minFractId){
  _inputSeqs.insert( seqs->begin(), seqs->end() );
}


AssemblyJobSubset::~AssemblyJobSubset(){
}


void AssemblyJobSubset::makeShallow(bool shallowDelete){
  if (_wasRun){
    throw AssemblyException::CallingError("AJSubset::makeShallow can't be called after the job was run.");
  }
  set<ScoredSeq*> shallowCopies;
  // buffer all input, then copy
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    (*it)->bottomBuffer();
  }
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    shallowCopies.insert( (*it)->shallowCopy() );
    (*it)->deepUnbuffer();
  }
  if ( shallowDelete){
    for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
      delete (*it);
    }
  }
  // replace the input contents with the shallow copies
  _inputSeqs.clear();
  _inputSeqs.insert( shallowCopies.begin(), shallowCopies.end() );
}


bool AssemblyJobSubset::wasRun(){ return _wasRun; }

void AssemblyJobSubset::allInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}

void AssemblyJobSubset::remainingInputSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobSubset::remainingInputSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobSubset::remainingInputSeqs; seqs were already removed.");
  }
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
}

void AssemblyJobSubset::discardedInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}

void AssemblyJobSubset::newlyAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobSubset::newlyAssembledSeqs; job hasn't run yet.");
  }
  if ( _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobSubset::newlyAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}

void AssemblyJobSubset::allAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobSubset::allAssembledSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved or _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobSubset::allAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}



void AssemblyJobSubset::OK(){
  if (! _wasRun ){
    if ( _novelSeqs.size() != 0 or _allAssembledSeqs.size() != 0 ){ //or _remainingSeqs.size() != 0 ){
      throw AssemblyException::LogicError("AssemblyJobSubset::OK; there should be no product seqs if job hasn't run.");
    }
    if ( _novelRemoved or _inputRemoved ){
      throw AssemblyException::LogicError("AssemblyJobSubset::OK; seqs can't have been removed if job hasn't run.");
    }
  }
}




bool AssemblyJobSubset::SortSsnByLength::operator() (ScoredSeqNormalized* ssnA, ScoredSeqNormalized* ssnB){
  return ssnA->size() < ssnB->size();
}



void AssemblyJobSubset::runJob(AssemblyJob::AssemblyThreadedness threadedness){
  if (! _wasRun){
    runJobMapping(threadedness);
    runJobCombine();
    _wasRun = true;
  }
}




void AssemblyJobSubset::runJobMappingGetAlignmentsHelper(ScoredSeqNormalized* shortSeq, ScoredSeqCollectionBwt* longSeqBwt, set<Alignment*>* shortIndexedAlSet){
  // since this is the short seq, it can only find LONGER seqs
  MatchSeqTest* matchSeqTest = new MatchSeqTestMinLength( shortSeq->size() + 1);
  //long lastPos = shortSeq->size() - 1; // this position should not be gapped in the short sequence

  if (_strandedness == AssemblyJob::doubleStranded){
    longSeqBwt->getMatches(shortIndexedAlSet, shortSeq, matchSeqTest, shortSeq->size(), ScoredSeqCollectionBwt::semiGlobal);
  } else {
    longSeqBwt->getMatches(shortIndexedAlSet, shortSeq, '+', matchSeqTest, shortSeq->size(), ScoredSeqCollectionBwt::semiGlobal);
  }
  delete matchSeqTest;
}



void AssemblyJobSubset::runJobMapping(AssemblyJob::AssemblyThreadedness threadedness){
  // set up short-seq-indexed data structures
  long numInput = _inputSeqs.size();
  // fill these in during the step 1 loop
  ScoredSeqNormalized** shortSeqArray = new ScoredSeqNormalized*[ numInput ];
  set<Alignment*>** shortIndexedAlSets = new set<Alignment*>*[ numInput ];

  // STEP 1: create a BWT collection from all the sufficient-length sequences
  // and set up collections that will allow me to keep track of all the match
  // combinations even as I start consolidating sequences
  set<ScoredSeq*> inputForBwt; // cast differently for SSCBwt input
  long shortIndex = 0;
  for (set<ScoredSeq*>::iterator inputIt = _inputSeqs.begin(); inputIt != _inputSeqs.end(); ++inputIt){
    ScoredSeqNormalized* newSsn = new ScoredSeqNormalized( (*inputIt) );
    _initialCarriers.insert( newSsn );
    inputForBwt.insert( newSsn );
    shortSeqArray[shortIndex] = newSsn;
    shortIndexedAlSets[shortIndex] = new set<Alignment*>;
    shortIndex++;
  }
  // even though there is not a pre-specified min overlap, the alignments will all be run using
  // an input-specified min overlap, so penalize edge gaps is false
  ScoredSeqCollectionBwt* longSeqBwt = new ScoredSeqCollectionBwt( &inputForBwt, _minFractId, false );
  longSeqBwt->disableEdgeScaling();

  // STEP 2: query all the sufficient-length sequences to that collection
  // make a threading array of the seq test
  MatchSeqTest** matchTestArray = new MatchSeqTest*[ numInput+1 ];
  long* minOvlArray = new long[ numInput+1 ];
  ScoredSeq** recastSeqArray = new ScoredSeq*[ numInput+1 ];
  for (long n = 0; n < numInput; ++n){
    matchTestArray[n] = new MatchSeqTestMinLength( shortSeqArray[n]->size() + 1);
    minOvlArray[n] = shortSeqArray[n]->size();
    recastSeqArray[n] = shortSeqArray[n];
  }

  // THIS CAN BE THREADED
  ScoredSeqCollectionBwt::AlignmentThreadedness alThread;
  if (threadedness == AssemblyJob::notThreaded){ alThread = ScoredSeqCollectionBwt::notThreaded; }
  else { alThread = ScoredSeqCollectionBwt::threaded; }

  if (_strandedness == AssemblyJob::doubleStranded){
    longSeqBwt->getMatches(numInput, shortIndexedAlSets, recastSeqArray, matchTestArray, minOvlArray, alThread, ScoredSeqCollectionBwt::semiGlobal);
  } else {
    longSeqBwt->getMatches(numInput, shortIndexedAlSets, recastSeqArray, '+', matchTestArray, minOvlArray, alThread, ScoredSeqCollectionBwt::semiGlobal);
  }

  for (long n = 0; n < numInput; ++n){ delete matchTestArray[n]; }
  delete [] matchTestArray;
  delete [] minOvlArray;
  delete [] recastSeqArray;

  // now make sure everything is organized properly in a NON-THREADED loop
  // DO NOT delete non-empty array elements because those sets are now values in _shortToAlignments
  for (long n = 0; n < numInput; ++n){
    if (shortIndexedAlSets[n]->size() == 0){ delete shortIndexedAlSets[n]; }
    else {
      shortSeqArray[n]->addCounts( shortIndexedAlSets[n]->size() );
      // this set becomes the value of the new short table entry
      _shortToAlignments.insert( pair<ScoredSeqNormalized*, set<Alignment*>*>( shortSeqArray[n], shortIndexedAlSets[n] ) );
      // now I must create or add to the long table entry
      for (set<Alignment*>::iterator alIt = shortIndexedAlSets[n]->begin(); alIt != shortIndexedAlSets[n]->end(); ++alIt){
	Alignment* al = *alIt;
	ScoredSeqNormalized* longSeq = dynamic_cast<ScoredSeqNormalized*>( al->seqB() );
	map<ScoredSeqNormalized*, set<Alignment*>* >::iterator findLongIt = _longToAlignments.find( longSeq );
	if (findLongIt == _longToAlignments.end()){
	  set<Alignment*>* newAlSet = new set<Alignment*>;
	  newAlSet->insert( al );
	  _longToAlignments.insert( pair<ScoredSeqNormalized*, set<Alignment*>*>( longSeq, newAlSet ) );
	} else { findLongIt->second->insert( al ); }
      }
    }
  }
  delete [] shortSeqArray;
  delete [] shortIndexedAlSets;
  delete longSeqBwt;

  // any key of either table will be consumed
  for (map<ScoredSeqNormalized*, set<Alignment*>* >::iterator it = _shortToAlignments.begin(); it != _shortToAlignments.end(); ++it){
    _inputSeqs.erase( it->first->getNested() );
    _usedSeqs.insert( it->first->getNested() );
    //_initialCarriers.insert( it->first );
  }
  for (map<ScoredSeqNormalized*, set<Alignment*>* >::iterator it = _longToAlignments.begin(); it != _longToAlignments.end(); ++it){
    _inputSeqs.erase( it->first->getNested() );
    _usedSeqs.insert( it->first->getNested() );
    _initialLongs.push_back( it->first );
    //_initialCarriers.insert( it->first );
  }
  SortSsnByLength ssnSorter; // sort by seq length
  sort( _initialLongs.begin(), _initialLongs.end(), ssnSorter );
}




void AssemblyJobSubset::runJobCombine(){

  for (vector<ScoredSeqNormalized*>::iterator oldSeqIt = _initialLongs.begin(); oldSeqIt != _initialLongs.end(); ++oldSeqIt){
    ScoredSeqNormalized* oldLong = (*oldSeqIt);

    // the long set is guaranteed to exist; get it and delete the map entry
    map<ScoredSeqNormalized*, set<Alignment*>* >::iterator longSetIt = _longToAlignments.find(oldLong);
    if (longSetIt == _longToAlignments.end()){
      throw AssemblyException::LogicError("long key, which is guaranteed to exist, was not found 1st pass");
    }
    set<Alignment*>* longSet = longSetIt->second;
    _longToAlignments.erase(longSetIt);

    // combine the sequences to yield a new normalized seq
    ScoredSeqNormalized* newSeq = combineHelper(longSet, oldLong);
    delete longSet;
    if (oldLong == newSeq){ throw AssemblyException::LogicError("AJSubset: alignment set cannot be empty."); }


    // if the sequence does not exist in the short set, that means the end of the line has been
    // reached and the product can be added to the output set
    map<ScoredSeqNormalized*, set<Alignment*>* >::iterator shortSetIt = _shortToAlignments.find(oldLong);
    set<Alignment*>* shortSet;
    if (shortSetIt == _shortToAlignments.end()){
      _novelSeqs.insert(newSeq->shallowCopy());
      newSeq->deepDelete();
    } else {
      // deletion must be postponed since these will be re-combined as short sequences
      _transientCarriers.insert(newSeq);

      shortSet = shortSetIt->second;
      _shortToAlignments.erase(shortSetIt);
      for (set<Alignment*>::iterator shortAlIt = shortSet->begin(); shortAlIt != shortSet->end(); ++shortAlIt){
	Alignment* shortAl = (*shortAlIt);
	ScoredSeqNormalized* localLong = dynamic_cast<ScoredSeqNormalized*>( shortAl->seqB() );
	shortAl->seqReplace(newSeq, localLong);
	// guaranteed to exist
	map<ScoredSeqNormalized*, set<Alignment*>* >::iterator localLongSetIt = _longToAlignments.find(localLong);
	if (localLongSetIt == _longToAlignments.end()){
	  throw AssemblyException::LogicError("long key, which is guaranteed to exist, was not found 2nd pass");
	}
      }
      delete shortSet;
    }
  }

  // short-to-long was not cleared during the collapse
  for (map<ScoredSeqNormalized*, set<Alignment*>*>::iterator it = _shortToAlignments.begin(); it != _shortToAlignments.end(); ++it){
    delete it->second;
  }
  // erase the wrappers for the initial sequences
  for (set<ScoredSeqNormalized*>::iterator trashIt = _initialCarriers.begin(); trashIt != _initialCarriers.end(); ++trashIt){
    delete (*trashIt);
  }
  // deep-delete the transient sequences
  for (set<ScoredSeqNormalized*>::iterator trashIt = _transientCarriers.begin(); trashIt != _transientCarriers.end(); ++trashIt){
    (*trashIt)->deepDelete();
  }
}




void AssemblyJobSubset::printAlignNw(Alignment * al){
  if ( al->isNull() ){
    throw AssemblyException::ArgError("in printAlignNw, alignment may not be null.");
  }
  ScoredSeq * seqA = al->seqA();
  ScoredSeq * seqB = al->seqB();
  char sense = al->orientation();
  long posA = 0;
  long posB = 0;
  long maxA = seqA->size();
  long maxB = seqB->size();

  string stringA = "";
  string stringB = "";
  string matchLines = "";

  while ( posA < maxA or posB < maxB ){

    if ( posA < maxA and al->isGapped(posA,seqA) ){
      stringA += seqA->nucAtPosPlus(posA);
      stringB += '-';
      matchLines += ' ';
      posA++;
    } else if ( posB < maxB and al->isGapped(posB,seqB) ){
      stringA += '-';
      stringB += seqB->nucAtPosition(posB,sense);
      matchLines += ' ';
      posB++;
    } else {
      char nucA = seqA->nucAtPosPlus(posA);
      char nucB = seqB->nucAtPosition(posB,sense);
      stringA += nucA;
      stringB += nucB;
      if ( nucA == nucB ){ matchLines += '|'; }
      else { matchLines += ' '; }
      posA++;
      posB++;
    }

  }

  cout << stringA << endl;
  cout << matchLines << endl;
  cout << stringB << endl;
}



ScoredSeqNormalized* AssemblyJobSubset::combineHelper(set<Alignment*>* alSet, ScoredSeqNormalized* oldSeq){
  if (alSet->size() == 0){ return oldSeq; }
  else {

    // this will get updated; I am wrapping it so that I can iteratively delete it without the original being deleted
    ScoredSeq* currentInner = ScoredSeqFlip::getFlip(oldSeq->getNested(), '+');

    for (set<Alignment*>::iterator alIt = alSet->begin(); alIt != alSet->end(); ++alIt){
      Alignment* al = (*alIt);
      // first, replace the seqB in the alignment with the current modification to seqB
      al->seqReplace( al->seqA(), currentInner );
      // storage so i can delete it after replacing it
      ScoredSeq* priorInner = currentInner;

      // second, combine the alignment, flipping it over to match the sense of seqB and updating currentInner
      if (al->orientation() == '+'){
	currentInner = al->combine();
      } else if (al->orientation() == '-'){
	ScoredSeq* seqaFlip = ScoredSeqFlip::getFlip(al->seqA(),'-');
	Alignment* alFlip = al->copyRcSeqA(seqaFlip);
	currentInner = alFlip->combine();
	delete seqaFlip;
	delete alFlip;
      } else {
	throw AssemblyException::ArgError("AJSubset::combineHelper, sense of alignment is an invalid char");
      }

      // third, clean up
      delete al;
      delete priorInner;
    }

    ScoredSeqNormalized* carrier = new ScoredSeqNormalized( currentInner );
    carrier->addCounts( oldSeq->getCount() );
    return carrier;
  }
}


#endif
